@@ -1,7 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2004-2005 Imendio AB
- * Copyright (C) 2004 Martyn Russell <mr@gnome.org>
+ * Copyright (C) 2004-2006 Imendio AB
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -43,19 +42,26 @@
 #include "gossip-jabber-private.h"
 
 #define DEBUG_MSG(x)
-/* #define DEBUG_MSG(args) g_printerr args ; g_printerr ("\n");  */
+/* #define DEBUG_MSG(args) g_printerr args ; g_printerr ("\n");   */
 
 #define GET_PRIV(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GOSSIP_TYPE_JABBER, GossipJabberPriv))
 
+/* Common XMPP XML namespaces we use. */
 #define XMPP_VERSION_XMLNS  "jabber:iq:version"
 #define XMPP_ROSTER_XMLNS   "jabber:iq:roster"
 #define XMPP_REGISTER_XMLNS "jabber:iq:register"
 
-/* 3.5 minutes, we use this because if the port is just wrong then it
-   will timeout before then with that error */
+/* We use 3.5 minutes because if the port is just wrong then it
+ * will timeout before then with that error.
+ */
 #define CONNECT_TIMEOUT     210 
 
-
+/* This is the timeout we will accept the user to be composing for
+ * before we assume it is stuck and the server has failed to tell us
+ * the user has stopped composing.
+ */
+#define COMPOSING_TIMEOUT   45
+  
 struct _GossipJabberPriv {
 	LmConnection          *connection;
 	
@@ -66,27 +72,29 @@ struct _GossipJabberPriv {
 
 	GHashTable            *contacts;
 
-	/* cancel registration attempt */
+	/* Cancel registration attempt */
 	gboolean               register_cancel;
 
-	/* connection details */
+	/* Connection details */
 	guint                  connection_timeout_id;
 	gboolean               disconnect_request;
 
-	/* extended parts */
+	/* Extended parts */
 	GossipJabberChatrooms *chatrooms;
 	GossipJabberFTs       *fts;
 
-	/* used to hold a list of composing message ids, this is so we
-	   can send the cancelation to the last message id */
+	/* Used to hold a list of composing message ids, this is so we
+	 * can send the cancelation to the last message id. 
+	 */
 	GHashTable            *composing_ids;
+	GHashTable            *composing_timeouts;
+	GHashTable            *composing_requests;
 
-	/* transport stuff... is this in the right place? */
+	/* Transport stuff... is this in the right place? */
 	GossipTransportAccountList *account_list;
 	LmMessageHandler      *subscription_handler;
 };
 
-
 typedef struct {
 	GossipJabber     *jabber;
 	LmMessageHandler *message_handler;
@@ -95,19 +103,17 @@ typedef struct {
 	gpointer          user_data;
 } RegisterData;
 
-
 typedef struct {
 	GossipJabber     *jabber;
 	gchar            *group;
 	gchar            *new_name;
 } RenameGroupData;
 
-
 typedef struct {
 	GossipJabber     *jabber;
 	GossipContact    *contact;
-} VCardData;
-
+	gpointer          user_data;
+} JabberData;
 
 static void             gossip_jabber_class_init            (GossipJabberClass       *klass);
 static void             gossip_jabber_init                  (GossipJabber            *jabber);
@@ -150,6 +156,7 @@ static void             jabber_register_cancel              (GossipProtocol
 static void             jabber_register_connection_open_cb  (LmConnection            *connection,
 							     gboolean                 result,
 							     RegisterData            *ra);
+static gboolean         jabber_composing_timeout_cb         (JabberData              *data);
 static LmHandlerResult  jabber_register_message_handler     (LmMessageHandler        *handler,
 							     LmConnection            *conn,
 							     LmMessage               *m,
@@ -196,7 +203,7 @@ static void             jabber_contact_vcard                (GossipJabber
 							     GossipContact           *contact);
 static void             jabber_contact_vcard_cb             (GossipResult             result,
 							     GossipVCard             *vcard,
-							     VCardData               *vd);
+							     JabberData              *data);
 static void             jabber_group_rename                 (GossipProtocol          *protocol,
 							     const gchar             *group,
 							     const gchar             *new_name);
@@ -246,7 +253,6 @@ static void             jabber_request_roster               (GossipJabber
 static void             jabber_request_unknown              (GossipJabber            *jabber,
 							     LmMessage               *m);
 
-
 /* chatrooms */
 static void             jabber_chatroom_init                (GossipChatroomProviderIface *iface);
 static GossipChatroomId jabber_chatroom_join                (GossipChatroomProvider  *provider,
@@ -266,20 +272,21 @@ static void             jabber_chatroom_change_nick         (GossipChatroomProvi
 							     const gchar             *new_nick);
 static void             jabber_chatroom_leave               (GossipChatroomProvider  *provider,
 							     GossipChatroomId         id);
-static const gchar *    jabber_chatroom_get_room_name       (GossipChatroomProvider  *provider,
+static GossipChatroom * jabber_chatroom_find                (GossipChatroomProvider  *provider,
 							     GossipChatroomId         id);
 static void             jabber_chatroom_invite              (GossipChatroomProvider  *provider,
 							     GossipChatroomId         id,
-							     const gchar             *contact_id,
-							     const gchar             *invite);
+							     GossipContact           *contact,
+							     const gchar             *reason);
 static void             jabber_chatroom_invite_accept       (GossipChatroomProvider  *provider,
 							     GossipChatroomJoinCb     callback,
-							     const gchar             *nickname,
-							     const gchar             *invite_id);
+							     GossipChatroomInvite    *invite,
+							     const gchar             *nickname);
+static void             jabber_chatroom_invite_decline      (GossipChatroomProvider  *provider,
+							     GossipChatroomInvite    *invite,
+							     const gchar             *reason);
 static GList *          jabber_chatroom_get_rooms           (GossipChatroomProvider  *provider);
 
-
-
 /* fts */
 static void             jabber_ft_init                      (GossipFTProviderIface   *iface);
 static GossipFTId       jabber_ft_send                      (GossipFTProvider        *provider,
@@ -292,18 +299,20 @@ static void             jabber_ft_accept                    (GossipFTProvider
 static void             jabber_ft_decline                   (GossipFTProvider        *provider,
 							     GossipFTId               id);
 
-
+/* misc */
+static JabberData *     jabber_data_new                     (GossipJabber            *jabber,
+							     GossipContact           *contact,
+							     gpointer                 user_data);
+static void             jabber_data_free                    (JabberData              *data);
 
 extern GConfClient *gconf_client;
 
-
 G_DEFINE_TYPE_WITH_CODE (GossipJabber, gossip_jabber, GOSSIP_TYPE_PROTOCOL,
 			 G_IMPLEMENT_INTERFACE (GOSSIP_TYPE_CHATROOM_PROVIDER,
 						jabber_chatroom_init);
 			 G_IMPLEMENT_INTERFACE (GOSSIP_TYPE_FT_PROVIDER,
 						jabber_ft_init));
 
-
 static void
 gossip_jabber_class_init (GossipJabberClass *klass)
 {
@@ -362,6 +371,18 @@ gossip_jabber_init (GossipJabber *jabber)
 				       gossip_contact_equal,
 				       (GDestroyNotify) g_object_unref,
 				       (GDestroyNotify) g_free);
+
+	priv->composing_timeouts = 
+		g_hash_table_new_full (gossip_contact_hash,
+				       gossip_contact_equal,
+				       (GDestroyNotify) g_object_unref,
+				       (GDestroyNotify) jabber_data_free);
+
+	priv->composing_requests = 
+		g_hash_table_new_full (gossip_contact_hash,
+				       gossip_contact_equal,
+				       (GDestroyNotify) g_object_unref,
+				       NULL);
 }
 	
 static void
@@ -388,6 +409,8 @@ jabber_finalize (GObject *obj)
 	gossip_jabber_ft_finalize (priv->fts);
 
  	g_hash_table_destroy (priv->composing_ids);
+ 	g_hash_table_destroy (priv->composing_timeouts);
+ 	g_hash_table_destroy (priv->composing_requests);
 
 	if (priv->connection_timeout_id != 0) {
 		g_source_remove (priv->connection_timeout_id);
@@ -407,7 +430,7 @@ jabber_setup (GossipProtocol *protocol,
  	GossipJabber     *jabber;
  	GossipJabberPriv *priv;
  	LmMessageHandler *handler;
-	const gchar      *id;
+	gchar            *id;
 	GossipJID        *jid;
 	
  	g_return_if_fail (GOSSIP_IS_JABBER (protocol));
@@ -420,14 +443,17 @@ jabber_setup (GossipProtocol *protocol,
  
  	priv->contact = gossip_contact_new (GOSSIP_CONTACT_TYPE_USER,
 					    priv->account);
- 
-	id = gossip_account_get_id (account);
+	
+	id = g_strdup_printf ("%s/%s", 
+			      gossip_account_get_id (account),
+			      gossip_account_get_resource (account));
 	jid = gossip_jid_new (id);
 
 	g_object_set (priv->contact,
 		      "id", id,          /* we need FULL JID here */
 		      "name", id,
 		      NULL);
+	g_free (id);
 
 	priv->connection = lm_connection_new (gossip_account_get_server (account));
 
@@ -614,6 +640,24 @@ jabber_logout_contact_foreach (gpointer       key,
 			       GossipContact *contact,
 			       GossipJabber  *jabber)
 {
+	GList          *presences;
+	GList          *l;
+	GossipPresence *presence;
+
+	/* Set each contact to be offline, since they effectively are
+	 * now we don't know.
+	 */
+	presences = gossip_contact_get_presence_list (contact);
+	for (l = presences; l; l = l->next) {
+		presence = l->data;
+
+		if (!presence) {
+			continue;
+		}
+
+		gossip_contact_remove_presence (contact, presence);
+	}
+
 	g_signal_emit_by_name (jabber, "contact-removed", contact);
 	return TRUE;
 }
@@ -670,31 +714,44 @@ jabber_connection_open_cb (LmConnection *connection,
 	
 	priv = GET_PRIV (jabber);
 
-	if (priv->connection_timeout_id != 0) {
-		g_source_remove (priv->connection_timeout_id);
-		priv->connection_timeout_id = 0;
-	}
-
 	if (priv->disconnect_request) {
-		/* this is so we go no further that way we don't issue
-		   warnings for connections we stopped ourselves */
+		/* This is so we go no further that way we don't issue
+		 * warnings for connections we stopped ourselves.
+		 */
+		jabber_logout (GOSSIP_PROTOCOL (jabber));
 		return;
 	}
 
 	if (result == FALSE) {
+		jabber_logout (GOSSIP_PROTOCOL (jabber));
 		jabber_error (GOSSIP_PROTOCOL (jabber), GOSSIP_PROTOCOL_NO_CONNECTION);
 		return;
 	}
 
 	DEBUG_MSG (("Protocol: Connection open!"));
 
+	if (priv->connection_timeout_id != 0) {
+		g_source_remove (priv->connection_timeout_id);
+		priv->connection_timeout_id = 0;
+	}
+
 	account = priv->account;
 	account_password = gossip_account_get_password (account);
 
 	if (!account_password || strlen (account_password) < 1) {
-		/* FIXME: Ask the user for the password */
+		DEBUG_MSG (("Protocol: Requesting password for:'%s'", 
+			    gossip_account_get_id (account)));
+
 		g_signal_emit_by_name (jabber, "get-password", 
 				       account, &password);
+
+		if (!password) {
+			DEBUG_MSG (("Protocol: Cancelled password request for:'%s'", 
+				    gossip_account_get_id (account)));
+
+			jabber_logout (GOSSIP_PROTOCOL (jabber));
+			return;
+		}
 	} else {
 		password = g_strdup (account_password);
 	}
@@ -704,8 +761,8 @@ jabber_connection_open_cb (LmConnection *connection,
 	DEBUG_MSG (("Protocol: Attempting to use JabberID:'%s'", jid_str));
 
 	id = gossip_jid_string_get_part_name (jid_str);
+	resource = gossip_account_get_resource (account);
 
-	resource = gossip_jid_string_get_part_resource (jid_str);
 	if (!resource) {
 		DEBUG_MSG (("Protocol: JabberID:'%s' is invalid, there is no resource.", jid_str));
 
@@ -737,6 +794,7 @@ jabber_connection_auth_cb (LmConnection *connection,
 	priv = GET_PRIV (jabber);
 	
 	if (result == FALSE) {
+		jabber_logout (GOSSIP_PROTOCOL (jabber));
 		jabber_error (GOSSIP_PROTOCOL (jabber), GOSSIP_PROTOCOL_AUTH_FAILED);
 		return;
 	}
@@ -810,14 +868,14 @@ jabber_disconnect_cb (LmConnection       *connection,
 		priv->connection_timeout_id = 0;
 	}
 
-	g_signal_emit_by_name (jabber, "logged-out", priv->account);
-
-	/* signal removal of each contact */
+	/* Signal removal of each contact */
 	if (priv->contacts) {
 		g_hash_table_foreach_remove (priv->contacts,
 					     (GHRFunc) jabber_logout_contact_foreach,
 					     jabber);
 	}
+
+	g_signal_emit_by_name (jabber, "logged-out", priv->account);
 }
 
 static LmSSLResponse 
@@ -981,6 +1039,8 @@ jabber_register_connection_open_cb (LmConnection *connection,
 	LmMessage        *m;
 	LmMessageNode    *node;
 	const gchar      *error_message = NULL;
+	const gchar      *id;
+	gchar            *username;
 	gboolean          ok = FALSE;
 
 	priv = GET_PRIV (ra->jabber);
@@ -1026,7 +1086,11 @@ jabber_register_connection_open_cb (LmConnection *connection,
 
         lm_message_node_set_attribute (node, "xmlns", XMPP_REGISTER_XMLNS);
 	
-	lm_message_node_add_child (node, "username", gossip_account_get_id (priv->account));
+	id = gossip_account_get_id (priv->account);
+	username = gossip_jid_string_get_part_name (id);
+	lm_message_node_add_child (node, "username", username);
+	g_free (username);
+
 	lm_message_node_add_child (node, "password", gossip_account_get_password (priv->account));
 
         ok = lm_connection_send_with_reply (connection, m, 
@@ -1176,7 +1240,7 @@ jabber_is_valid_username (GossipProtocol *protocol,
 	jabber = GOSSIP_JABBER (protocol);
 	priv = GET_PRIV (jabber);
 
-	return gossip_jid_string_is_valid (username, TRUE);
+	return gossip_jid_string_is_valid (username, FALSE);
 }
 
 static gboolean
@@ -1188,7 +1252,7 @@ jabber_is_ssl_supported (GossipProtocol *protocol)
 static const gchar *
 jabber_get_example_username (GossipProtocol *protocol)
 {
-	return "user@jabber.org/laptop";
+	return "user@jabber.org";
 }
 
 static gchar *
@@ -1266,7 +1330,7 @@ jabber_send_message (GossipProtocol *protocol,
 		jid_str = g_strdup (recipient_id);
 	}
 
-	DEBUG_MSG (("Protocol: Sending message to: '%s'", jid_str));
+	DEBUG_MSG (("Protocol: Sending message to:'%s'", jid_str));
 	
 	m = lm_message_new_with_sub_type (jid_str,
 					  LM_MESSAGE_TYPE_MESSAGE,
@@ -1275,8 +1339,9 @@ jabber_send_message (GossipProtocol *protocol,
 	lm_message_node_add_child (m->node, "body",
 				   gossip_message_get_body (message));
 
-	/* if we have had a request for composing then we send the
-	   other side composing details with every message */
+	/* If we have had a request for composing then we send the
+	 * other side composing details with every message 
+	 */
 	if (gossip_message_is_requesting_composing (message)) {
 		LmMessageNode *node;
 
@@ -1302,7 +1367,6 @@ jabber_send_composing (GossipProtocol *protocol,
 	LmMessageNode    *node;
 	const gchar      *id = NULL;
 	const gchar      *contact_id;
-	const gchar      *resource = NULL;
 	gchar            *jid_str;
 
 	g_return_if_fail (GOSSIP_IS_JABBER (protocol));
@@ -1311,20 +1375,17 @@ jabber_send_composing (GossipProtocol *protocol,
 	jabber = GOSSIP_JABBER (protocol);
 	priv = GET_PRIV (jabber);
 
+	if (!g_hash_table_lookup (priv->composing_requests, contact)) {
+		return;
+	}
+
 	contact_id = gossip_contact_get_id (contact);
 
 	DEBUG_MSG (("Protocol: Sending %s to contact:'%s'", 
 		   typing ? "composing" : "not composing",
 		   contact_id));
 	
-	/* FIXME: Create a full JID (with resource) and send to that */
-/* 	resource = gossip_message_get_explicit_resource (message); */
-
-	if (resource) {
-		jid_str = g_strdup_printf ("%s/%s", contact_id, resource);
-	} else {
-		jid_str = g_strdup (contact_id);
-	}
+	jid_str = g_strdup (contact_id);
 
 	m = lm_message_new_with_sub_type (jid_str,
 					  LM_MESSAGE_TYPE_MESSAGE,
@@ -1492,14 +1553,22 @@ jabber_contact_add (GossipProtocol *protocol,
                     const gchar    *group,
                     const gchar    *message)
 {
-        GossipJabber     *jabber;
-        GossipJabberPriv *priv;
-	LmMessage        *m;
-        LmMessageNode    *node;
+	LmMessage          *m;
+        LmMessageNode      *node;
+        GossipJabber       *jabber;
+        GossipJabberPriv   *priv;
+	GossipJID          *jid;
+	GossipContact      *contact;
+	GossipContactType   type;
+	GossipSubscription  subscription;
+	gboolean            ask = TRUE;
 
         jabber = GOSSIP_JABBER (protocol);
         priv = GET_PRIV (jabber);
 
+	jid = gossip_jid_new (id);
+	contact = g_hash_table_lookup (priv->contacts, gossip_jid_get_without_resource (jid));
+
         /* Request subscription */
         m = lm_message_new_with_sub_type (id, LM_MESSAGE_TYPE_PRESENCE,
                                           LM_MESSAGE_SUB_TYPE_SUBSCRIBE);
@@ -1507,25 +1576,50 @@ jabber_contact_add (GossipProtocol *protocol,
         lm_connection_send (priv->connection, m, NULL);
         lm_message_unref (m);
 
-        /* Add to roster */
-        m = lm_message_new_with_sub_type (NULL, LM_MESSAGE_TYPE_IQ,
-                                          LM_MESSAGE_SUB_TYPE_SET);
-        node = lm_message_node_add_child (m->node, "query", NULL);
-        lm_message_node_set_attributes (node,
-                                        "xmlns", XMPP_ROSTER_XMLNS, NULL);
-        node = lm_message_node_add_child (node, "item", NULL);
-        lm_message_node_set_attributes (node,
-                                        "jid", id,
-                                        "subscription", "none",
-                                        "ask", "subscribe",
-                                        "name", name,
-                                        NULL);
+	/* Add to roster */
+	m = lm_message_new_with_sub_type (NULL, LM_MESSAGE_TYPE_IQ,
+					  LM_MESSAGE_SUB_TYPE_SET);
+
+	node = lm_message_node_add_child (m->node, "query", NULL);
+	lm_message_node_set_attributes (node,
+					"xmlns", XMPP_ROSTER_XMLNS, NULL);
+
+	node = lm_message_node_add_child (node, "item", NULL);
+
+	/* Only ask for subscription if we already have NONE */
+	if (contact) {
+		type = gossip_contact_get_type (contact);
+		subscription = gossip_contact_get_subscription (contact);
+		
+		ask &= type == GOSSIP_CONTACT_TYPE_TEMPORARY;
+		ask &= subscription != GOSSIP_SUBSCRIPTION_BOTH;
+	} else {
+		ask = FALSE;
+	}
+
+	if (ask) {
+		lm_message_node_set_attributes (node,
+						"jid", gossip_jid_get_without_resource (jid),
+						"subscription", "none",
+						"ask", "subscribe",
+						"name", name,
+						NULL);
+	} else {
+		/* Make sure we set the name incase we updated it */
+		lm_message_node_set_attributes (node,
+						"jid", gossip_jid_get_without_resource (jid),
+						"name", name,
+						NULL);
+	}
+
         if (group && strcmp (group, "") != 0) {
                 lm_message_node_add_child (node, "group", group);
         }
+	
+	lm_connection_send (priv->connection, m, NULL);
+	lm_message_unref (m);
 
-        lm_connection_send (priv->connection, m, NULL);
-        lm_message_unref (m);
+	gossip_jid_unref (jid);
 }
 
 static void
@@ -1623,46 +1717,41 @@ jabber_contact_vcard (GossipJabber  *jabber,
 		      GossipContact *contact)
 {
 	GossipJabberPriv *priv;
-	VCardData        *vd;
+	JabberData       *data;
 	
 	priv = GET_PRIV (jabber);
 
-	vd = g_new0 (VCardData, 1);
-	
-	vd->jabber = g_object_ref (jabber);
-	vd->contact = g_object_ref (contact);
+	data = jabber_data_new (jabber, contact, NULL);
 	
 	gossip_jabber_vcard_get (priv->connection,
 				 gossip_contact_get_id (contact),
 				 (GossipVCardCallback) jabber_contact_vcard_cb, 
-				 vd, 
+				 data, 
 				 NULL);
 }
 
 static void
 jabber_contact_vcard_cb (GossipResult  result,
 			 GossipVCard  *vcard,
-			 VCardData    *vd)
+			 JabberData   *data)
 {
 	if (result == GOSSIP_RESULT_OK) {
 		gchar *name;
 
 		name = gossip_jabber_get_name_to_use 
-			(gossip_contact_get_id (vd->contact),
+			(gossip_contact_get_id (data->contact),
 			 gossip_vcard_get_nickname (vcard),
 			 gossip_vcard_get_name (vcard));
 		
-		gossip_contact_set_name (vd->contact, name);
+		gossip_contact_set_name (data->contact, name);
 		g_free (name);
 
-		g_signal_emit_by_name (vd->jabber, 
+		g_signal_emit_by_name (data->jabber, 
 				       "contact-updated", 
-				       vd->contact);
+				       data->contact);
 	}
 
-	g_object_unref (vd->jabber);
-	g_object_unref (vd->contact);
-	g_free (vd);
+	jabber_data_free (data);
 }
 
 static void
@@ -1969,84 +2058,192 @@ jabber_set_proxy (LmConnection *conn)
 	g_object_unref (gconf_client);
 }
 
+static gboolean
+jabber_composing_timeout_cb (JabberData *data)
+{
+	GossipJabberPriv *priv;
+
+	priv = GET_PRIV (data->jabber);
+
+	DEBUG_MSG (("Protocol: Contact:'%s' is NOT composing (timed out)",
+		    gossip_contact_get_id (data->contact)));
+
+	g_signal_emit_by_name (data->jabber, "composing", data->contact, FALSE);
+
+	g_hash_table_remove (priv->composing_timeouts, data->contact);
+
+	return FALSE;
+}
+
 static LmHandlerResult
 jabber_message_handler (LmMessageHandler *handler,
 			LmConnection     *conn,
 			LmMessage        *m,
 			GossipJabber     *jabber)
 {
-	GossipJabberPriv *priv;
-	GossipMessage    *message;
-	const gchar      *from_str;
-	GossipContact    *from;
-	const gchar      *thread = "";
-	const gchar      *body = "";
-	LmMessageNode    *node;
+	LmMessageNode        *node;
+	LmMessageSubType      sub_type;
+	GossipJabberPriv     *priv;
+	GossipMessage        *message;
+	const gchar          *from_str;
+	GossipContact        *from;
+	const gchar          *thread = NULL;
+	const gchar          *subject = NULL;
+	const gchar          *body = NULL;
+	GossipChatroomInvite *invite;
 
 	priv = GET_PRIV (jabber);
 	
-	DEBUG_MSG (("Protocol: New message from: %s", 
+	DEBUG_MSG (("Protocol: New message from:'%s'", 
 		   lm_message_node_get_attribute (m->node, "from")));
 
-	switch (lm_message_get_sub_type (m)) {
-	case LM_MESSAGE_SUB_TYPE_NOT_SET:
-	case LM_MESSAGE_SUB_TYPE_NORMAL:
-	case LM_MESSAGE_SUB_TYPE_CHAT:
-	case LM_MESSAGE_SUB_TYPE_HEADLINE: /* For now, fixes #120009 */
-		from_str = lm_message_node_get_attribute (m->node, "from");
-		from = gossip_jabber_get_contact_from_jid (jabber, 
-							   from_str, 
-							   NULL,
-							   TRUE);
-		
-		message = gossip_message_new (GOSSIP_MESSAGE_TYPE_NORMAL,
-					      priv->contact);
-		
-		gossip_message_set_sender (message, from);
-
-		/* if event, then we can ignore the rest */
-		if (gossip_jabber_get_message_is_event (m)) {
-			gboolean composing;
-
-			composing = gossip_jabber_get_message_is_composing (m);
-
-			g_signal_emit_by_name (jabber, "composing-event", 
-					       from, composing);
-			break;
-		}
-
-		node = lm_message_node_get_child (m->node, "body");
-		if (node) {
-			body = node->value;
-		}
-
-		node = lm_message_node_get_child (m->node, "thread");
-		if (node) {
-			thread = node->value;
-		}
-			
-		gossip_message_set_body (message, body);
-		gossip_message_set_thread (message, thread);
-
-		gossip_message_set_timestamp (message,
-					      gossip_jabber_get_message_timestamp (m));
-		
-		gossip_message_set_invite (message,
-					   gossip_jabber_get_message_conference (m));
-		
-		g_signal_emit_by_name (jabber, "new-message", message);
-		g_signal_emit_by_name (jabber, "composing-event", 
-				       from, FALSE);
-		g_object_unref (message);
-
-		break;
-	case LM_MESSAGE_SUB_TYPE_ERROR:
-		/* FIXME: Emit error */
-		break;
-	default:
-		break;
+	sub_type = lm_message_get_sub_type (m);
+	if (sub_type != LM_MESSAGE_SUB_TYPE_NOT_SET && 
+	    sub_type != LM_MESSAGE_SUB_TYPE_NORMAL && 
+	    sub_type != LM_MESSAGE_SUB_TYPE_CHAT && 
+	    sub_type != LM_MESSAGE_SUB_TYPE_HEADLINE) {
+		return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
 	}
 
+	from_str = lm_message_node_get_attribute (m->node, "from");
+	from = gossip_jabber_get_contact_from_jid (jabber, 
+						   from_str, 
+						   NULL,
+						   TRUE);
+	
+	if (gossip_jabber_get_message_is_event (m)) {
+		gboolean composing;
+		
+		composing = gossip_jabber_get_message_is_composing (m);
+		
+		DEBUG_MSG (("Protocol: Contact:'%s' %s",
+			    gossip_contact_get_id (from),
+			    composing ? "is composing" : "is NOT composing"));
+		
+		g_signal_emit_by_name (jabber, "composing", from, composing);
+		
+		if (composing) {
+			JabberData *data;
+			guint       id;
+			
+			data = g_hash_table_lookup (priv->composing_timeouts, from);
+			if (data) {
+				g_source_remove (GPOINTER_TO_UINT (data->user_data));
+				g_hash_table_remove (priv->composing_timeouts, from);
+			}
+			
+			data = jabber_data_new (jabber, from, NULL);
+			id = g_timeout_add (COMPOSING_TIMEOUT * 1000,
+					    (GSourceFunc) jabber_composing_timeout_cb,
+					    data);
+			data->user_data = GUINT_TO_POINTER (id);
+			g_hash_table_insert (priv->composing_timeouts, g_object_ref (from), data);
+		} else {
+			JabberData *data;
+
+			data = g_hash_table_lookup (priv->composing_timeouts, from);
+			if (data) {
+				g_source_remove (GPOINTER_TO_UINT (data->user_data));
+				g_hash_table_remove (priv->composing_timeouts, from);
+			}
+		}
+		
+		/* If event, then we can ignore the rest */
+		return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
+	} else {
+		gboolean wants_composing;
+		
+		wants_composing = gossip_jabber_get_message_is_composing (m);
+		
+		if (wants_composing) {
+			g_hash_table_insert (priv->composing_requests, 
+					     g_object_ref (from),
+					     GINT_TO_POINTER (TRUE));
+		} else {
+			g_hash_table_remove (priv->composing_requests, 
+					     g_object_ref (from));
+		}
+		
+		DEBUG_MSG (("Protocol: Contact:'%s' %s composing info...",
+			    gossip_contact_get_id (from),
+			    wants_composing ? "wants" : "does NOT want"));
+	}
+	
+	node = lm_message_node_get_child (m->node, "subject");
+	if (node) {
+		subject = node->value;
+	}
+
+	invite = gossip_jabber_get_message_conference (jabber, m);
+	if (invite) {
+		GossipContact *invitor;
+		
+		invitor = gossip_chatroom_invite_get_invitor (invite);
+		
+		DEBUG_MSG (("Protocol: Chat room invitiation from:'%s' for room:'%s', reason:'%s'", 
+			    gossip_contact_get_id (invitor), 
+			    gossip_chatroom_invite_get_id (invite), 
+			    gossip_chatroom_invite_get_reason (invite)));
+		
+		/* We set the from to be the person that invited you,
+		 * not the chatroom that sent the request on their
+		 * behalf.
+		 */
+		from = invitor;
+		
+		/* Make sure we have some sort of body for
+		 * invitations, since it is not necessary but should
+		 * really exist.
+		 */
+		body = "";
+	}
+	
+	node = lm_message_node_get_child (m->node, "body");
+	if (node) {
+		body = node->value;
+	}
+
+	if (!body && !invite) {
+		/* If no body to the message, we ignore it since it
+		 * has no purpose now (see fixes #309912). 
+		 */
+
+		DEBUG_MSG (("Protocol: Dropping new message, no <body> element"));
+		return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
+	}
+	
+	node = lm_message_node_get_child (m->node, "thread");
+	if (node) {
+		thread = node->value;
+	}
+	
+	message = gossip_message_new (GOSSIP_MESSAGE_TYPE_NORMAL,
+				      priv->contact);
+
+	gossip_message_set_sender (message, from);
+	gossip_message_set_body (message, body);
+	
+	if (subject) {
+		gossip_message_set_subject (message, subject);
+	}
+	
+	if (thread) {
+		gossip_message_set_thread (message, thread);
+	}
+
+	if (invite) {
+		gossip_message_set_invite (message, invite); 
+		gossip_chatroom_invite_unref (invite);
+	}
+	
+	gossip_message_set_timestamp (message,
+				      gossip_jabber_get_message_timestamp (m));
+	
+	g_signal_emit_by_name (jabber, "new-message", message);
+	g_signal_emit_by_name (jabber, "composing", from, FALSE);
+	
+	g_object_unref (message);
+	
 	return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
 }
 
@@ -2065,7 +2262,7 @@ jabber_presence_handler (LmMessageHandler *handler,
 	priv = GET_PRIV (jabber);
 
 	from = lm_message_node_get_attribute (m->node, "from");
-        DEBUG_MSG (("Protocol: New presence from: %s", 
+        DEBUG_MSG (("Protocol: New presence from:'%s'", 
 		   lm_message_node_get_attribute (m->node, "from")));
 
 	if (gossip_jabber_chatrooms_get_jid_is_chatroom (priv->chatrooms,
@@ -2086,7 +2283,7 @@ jabber_presence_handler (LmMessageHandler *handler,
 
 		return LM_HANDLER_RESULT_REMOVE_MESSAGE;
 	} else if (strcmp (type, "subscribed") == 0) {
-		/* Handle this? */
+		/* FIXME: Handle this? */
 		return LM_HANDLER_RESULT_REMOVE_MESSAGE;
 	}
 	
@@ -2326,7 +2523,7 @@ jabber_request_roster (GossipJabber *jabber,
 
 		g_object_set (contact, "type", GOSSIP_CONTACT_TYPE_CONTACTLIST, NULL);
 
-		/* groups */
+		/* Groups */
 		for (subnode = node->children; subnode; subnode = subnode->next) {
 			if (strcmp (subnode->name, "group") != 0) {
 				continue;
@@ -2342,19 +2539,16 @@ jabber_request_roster (GossipJabber *jabber,
 			gossip_contact_set_groups (contact, groups);
 		}
 
-		/* subscription */
+		/* Subscription */
 		subscription = lm_message_node_get_attribute (node, "subscription");
 		if (contact && subscription) {
 			GossipSubscription type;
 
 			if (strcmp (subscription, "remove") == 0) {
-				g_object_ref (contact);
-				g_hash_table_remove (priv->contacts, 
-						     gossip_contact_get_id (contact));
-				
 				g_signal_emit_by_name (jabber,
 						       "contact-removed", contact);
-				g_object_unref (contact);
+				g_hash_table_remove (priv->contacts, 
+						     gossip_contact_get_id (contact));
 				continue;
 			} else if (strcmp (subscription, "both") == 0) {
 				type = GOSSIP_SUBSCRIPTION_BOTH;
@@ -2369,7 +2563,7 @@ jabber_request_roster (GossipJabber *jabber,
 			gossip_contact_set_subscription (contact, type);
 		} 
 		
-		/* find out if any thing has updated */
+		/* Find out if any thing has updated */
 		name_updated = groups_updated = FALSE;
 
 		name = lm_message_node_get_attribute (node, "name");
@@ -2385,7 +2579,7 @@ jabber_request_roster (GossipJabber *jabber,
 		}
 
 		if (new_groups) {
-			/* does not get free'd */
+			/* Does not get free'd */
 			groups_updated = gossip_contact_set_groups (contact, new_groups);
 		}
 
@@ -2458,9 +2652,10 @@ jabber_chatroom_init (GossipChatroomProviderIface *iface)
 	iface->change_topic    = jabber_chatroom_change_topic;
 	iface->change_nick     = jabber_chatroom_change_nick;
 	iface->leave           = jabber_chatroom_leave;
-	iface->get_room_name   = jabber_chatroom_get_room_name;
+	iface->find            = jabber_chatroom_find;
 	iface->invite          = jabber_chatroom_invite;
 	iface->invite_accept   = jabber_chatroom_invite_accept;
+	iface->invite_decline  = jabber_chatroom_invite_decline;
 	iface->get_rooms       = jabber_chatroom_get_rooms;
 }
 
@@ -2562,9 +2757,9 @@ jabber_chatroom_leave (GossipChatroomProvider *provider,
 	gossip_jabber_chatrooms_leave (priv->chatrooms, id);
 }
 
-static const gchar *
-jabber_chatroom_get_room_name (GossipChatroomProvider *provider,
-			       GossipChatroomId        id)
+static GossipChatroom *
+jabber_chatroom_find (GossipChatroomProvider *provider,
+		      GossipChatroomId        id)
 {
 	GossipJabber     *jabber;
 	GossipJabberPriv *priv;
@@ -2574,31 +2769,32 @@ jabber_chatroom_get_room_name (GossipChatroomProvider *provider,
 	jabber = GOSSIP_JABBER (provider);
 	priv = GET_PRIV (jabber);
 
-	return gossip_jabber_chatrooms_get_room_name (priv->chatrooms, id);
+	return gossip_jabber_chatrooms_find (priv->chatrooms, id);
 }
 
 static void
 jabber_chatroom_invite (GossipChatroomProvider *provider,
 			GossipChatroomId        id,
-			const gchar            *contact_id,
-			const gchar            *invite)
+			GossipContact          *contact,
+			const gchar            *reason)
 {
 	GossipJabber     *jabber;
 	GossipJabberPriv *priv;
 
 	g_return_if_fail (GOSSIP_IS_JABBER (provider));
+	g_return_if_fail (GOSSIP_IS_CONTACT (contact));
 
 	jabber = GOSSIP_JABBER (provider);
 	priv = GET_PRIV (jabber);
 
-	gossip_jabber_chatrooms_invite (priv->chatrooms, id, contact_id, invite);
+	gossip_jabber_chatrooms_invite (priv->chatrooms, id, contact, reason);
 }
 
 static void
 jabber_chatroom_invite_accept (GossipChatroomProvider *provider,
 			       GossipChatroomJoinCb    callback,
-			       const gchar            *nickname,
-			       const gchar            *invite_id)
+			       GossipChatroomInvite   *invite,
+			       const gchar            *nickname)
 {
 	GossipJabber     *jabber;
 	GossipJabberPriv *priv;
@@ -2610,8 +2806,26 @@ jabber_chatroom_invite_accept (GossipChatroomProvider *provider,
 
 	gossip_jabber_chatrooms_invite_accept (priv->chatrooms, 
 					       callback, 
-					       nickname, 
-					       invite_id);
+					       invite,
+					       nickname);
+}
+
+static void
+jabber_chatroom_invite_decline (GossipChatroomProvider *provider,
+				GossipChatroomInvite   *invite,
+				const gchar            *reason)
+{
+	GossipJabber     *jabber;
+	GossipJabberPriv *priv;
+
+	g_return_if_fail (GOSSIP_IS_JABBER (provider));
+
+	jabber = GOSSIP_JABBER (provider);
+	priv = GET_PRIV (jabber);
+
+	gossip_jabber_chatrooms_invite_decline (priv->chatrooms, 
+						invite,
+						reason);
 }
 
 static GList *
@@ -2707,8 +2921,50 @@ jabber_ft_decline (GossipFTProvider *provider,
 	return gossip_jabber_ft_decline (priv->fts, id);
 }
 
+/*
+ * Misc
+ */
+
+static JabberData *
+jabber_data_new (GossipJabber  *jabber,
+		 GossipContact *contact,
+		 gpointer       user_data)
+{
+	JabberData *data;
+
+	g_return_val_if_fail (GOSSIP_IS_JABBER (jabber), NULL);
+	g_return_val_if_fail (GOSSIP_IS_CONTACT (contact), NULL);
+	
+	data = g_new0 (JabberData, 1);
+	
+	data->jabber = g_object_ref (jabber);
+	data->contact = g_object_ref (contact);
+
+	data->user_data = user_data;
+
+	return data;
+}
+
+static void
+jabber_data_free (JabberData *data)
+{
+	if (!data) {
+		return;
+	}
+
+	if (data->jabber) {
+		g_object_unref (data->jabber);
+	}
+
+	if (data->contact) {
+		g_object_unref (data->contact);
+	}
+
+	g_free (data);
+}
+
 /* 
- * external functions 
+ * External functions 
  */ 
 
 GossipAccount *
@@ -2749,13 +3005,12 @@ gossip_jabber_get_contact_from_jid (GossipJabber *jabber,
 	priv = GET_PRIV (jabber);
 
 	jid = gossip_jid_new (jid_str);
-	DEBUG_MSG (("Protocol: Get contact: %s", gossip_jid_get_full (jid)));
-
+	
 	contact = g_hash_table_lookup (priv->contacts, 
 				       gossip_jid_get_without_resource (jid));
 
 	if (!contact) {
-		DEBUG_MSG (("Protocol: New contact"));
+		DEBUG_MSG (("Protocol: New contact:'%s'", gossip_jid_get_full (jid)));
 		contact = gossip_contact_new (GOSSIP_CONTACT_TYPE_TEMPORARY,
 					      priv->account);
 		gossip_contact_set_id (contact, 
@@ -2766,7 +3021,7 @@ gossip_jabber_get_contact_from_jid (GossipJabber *jabber,
 		tmp_new_item = TRUE;
 		
 		g_hash_table_insert (priv->contacts, 
-				     g_strdup (gossip_contact_get_id (contact)),
+				     g_strdup (gossip_jid_get_without_resource (jid)),
 				     g_object_ref (contact));
 		
 		if (get_vcard) {
@@ -2775,6 +3030,8 @@ gossip_jabber_get_contact_from_jid (GossipJabber *jabber,
 			 */
 			jabber_contact_vcard (jabber, contact);
 		}
+	} else {
+		DEBUG_MSG (("Protocol: Get contact:'%s'", gossip_jid_get_full (jid)));
 	}
 
 	gossip_jid_unref (jid);
@@ -2880,7 +3137,7 @@ gossip_jabber_send_unsubscribed (GossipJabber  *jabber,
 }
 
 /*
- * private functions 
+ * Private functions 
  */
 
 LmConnection *
@@ -2906,3 +3163,4 @@ gossip_jabber_get_fts (GossipJabber *jabber)
 	
 	return priv->fts;
 }
+
