@@ -40,7 +40,7 @@
 
 #include "eggtrayicon.h"
 
-#include "gossip-about.h"
+#include "gossip-about-dialog.h"
 #include "gossip-account-button.h"
 #include "gossip-accounts-dialog.h"
 #include "gossip-add-contact-window.h"
@@ -51,6 +51,7 @@
 #include "gossip-ft-window.h"
 #include "gossip-group-chat.h"
 #include "gossip-idle.h"
+#include "gossip-log-window.h"
 #include "gossip-marshal.h"
 #include "gossip-new-account-window.h"
 #include "gossip-new-message-dialog.h"
@@ -181,10 +182,12 @@ static void            app_connect_cb                       (GtkWidget
 							     GossipApp                *app);
 static void            app_disconnect_cb                    (GtkWidget                *window,
 							     GossipApp                *app);
-static void            app_join_group_chat_cb               (GtkWidget                *window,
-							     GossipApp                *app);
 static void            app_new_message_cb                   (GtkWidget                *widget,
 							     GossipApp                *app);
+static void            app_history_cb                       (GtkWidget                *window,
+							     GossipApp                *app);
+static void            app_join_group_chat_cb               (GtkWidget                *window,
+							     GossipApp                *app);
 static void            app_add_contact_cb                   (GtkWidget                *widget,
 							     GossipApp                *app);
 static void            app_show_offline_cb                  (GtkCheckMenuItem         *item,
@@ -207,6 +210,11 @@ static void            app_session_protocol_disconnected_cb (GossipSession
 							     GossipAccount            *account,
 							     GossipProtocol           *protocol,
 							     gpointer                  user_data);
+static void            app_session_protocol_error_cb        (GossipSession            *session,
+							     GossipProtocol           *protocol,
+							     GossipAccount            *account,
+							     GError                   *error,
+							     gpointer                  user_data);
 static gchar *         app_session_get_password_cb          (GossipSession            *session,
 							     GossipAccount            *account,
 							     gpointer                  user_data);
@@ -353,6 +361,9 @@ app_finalize (GObject *object)
 	g_signal_handlers_disconnect_by_func (priv->session,
 					      app_session_protocol_disconnected_cb, 
 					      NULL);
+	g_signal_handlers_disconnect_by_func (priv->session,
+					      app_session_protocol_error_cb, 
+					      NULL);
 	g_signal_handlers_disconnect_by_func (priv->session,
 					      app_session_get_password_cb, 
 					      NULL);
@@ -417,8 +428,7 @@ app_setup (GossipAccountManager *manager)
 	
 	priv->session = gossip_session_new (manager);
 
-	/* is this the best place for this, 
-	   perhaps in gossip-main.c? */
+	/* Is this the best place for this, perhaps in gossip-main.c? */
 #ifdef HAVE_DBUS
         gossip_dbus_init (priv->session);
 #endif
@@ -427,11 +437,11 @@ app_setup (GossipAccountManager *manager)
 	gossip_galago_init (priv->session);
 #endif
 
-	/* call init session dependent modules */
+	/* Call init session dependent modules */
 	gossip_subscription_dialog_init (priv->session);
 	gossip_ft_window_init (priv->session);
 
-	/* do we need first time start up druid? */
+	/* Do we need first time start up druid? */
 	if (gossip_new_account_window_is_needed ()) {
 		gossip_new_account_window_show (NULL);
 	}
@@ -448,6 +458,8 @@ app_setup (GossipAccountManager *manager)
 			    priv->event_manager);
 #endif
 
+	gossip_sound_init (priv->session);
+
 	g_signal_connect (manager, "account_added",
 			  G_CALLBACK (app_accounts_account_added_cb), 
 			  NULL);
@@ -464,6 +476,10 @@ app_setup (GossipAccountManager *manager)
 			  G_CALLBACK (app_session_protocol_disconnected_cb),
 			  NULL);
 
+	g_signal_connect (priv->session, "protocol-error",
+			  G_CALLBACK (app_session_protocol_error_cb),
+			  NULL);
+
 	g_signal_connect (priv->session, "get-password",
 			  G_CALLBACK (app_session_get_password_cb),
 			  NULL);
@@ -504,8 +520,9 @@ app_setup (GossipAccountManager *manager)
 			      "file_quit", "activate", app_quit_cb,
 			      "actions_connect", "activate", app_connect_cb,
  			      "actions_disconnect", "activate", app_disconnect_cb, 
-			      "actions_join_group_chat", "activate", app_join_group_chat_cb,
 			      "actions_new_message", "activate", app_new_message_cb,
+			      "actions_history", "activate", app_history_cb,
+			      "actions_join_group_chat", "activate", app_join_group_chat_cb,
 			      "actions_add_contact", "activate", app_add_contact_cb,
 			      "actions_show_offline", "toggled", app_show_offline_cb,
 			      "edit_accounts", "activate", app_accounts_cb,
@@ -830,10 +847,7 @@ app_connect_cb (GtkWidget *window,
 {
 	g_return_if_fail (GOSSIP_IS_APP (app));
 
-	/* currently we pass TRUE even though this is not a startup
-	   call to connect, but if we pass FALSE then *ALL* accounts
-	   will be connected. */
-	gossip_app_connect (FALSE);
+	gossip_app_connect (NULL, FALSE);
 }
 
 static void
@@ -845,13 +859,6 @@ app_disconnect_cb (GtkWidget *window,
 	app_disconnect ();
 }
 
-static void
-app_join_group_chat_cb (GtkWidget *window,
-			GossipApp *app)
-{
-	gossip_chatrooms_window_show (NULL, FALSE);
-}
-
 static void
 app_new_message_cb (GtkWidget *widget,
 		    GossipApp *app)
@@ -863,6 +870,24 @@ app_new_message_cb (GtkWidget *widget,
 	gossip_new_message_dialog_show (GTK_WINDOW (priv->window));
 }
 
+static void
+app_history_cb (GtkWidget *widget,
+		GossipApp *app)
+{
+	GossipAppPriv *priv;
+
+	priv = app->priv;
+
+	gossip_log_window_show (NULL, NULL);
+}
+
+static void
+app_join_group_chat_cb (GtkWidget *window,
+			GossipApp *app)
+{
+	gossip_chatrooms_window_show (NULL, FALSE);
+}
+
 static void
 app_add_contact_cb (GtkWidget *widget, 
 		    GossipApp *app)
@@ -928,16 +953,11 @@ static void
 app_about_cb (GtkWidget *window,
 	      GossipApp *app)
 {
-	static GtkWidget *about;
+	GossipAppPriv *priv;
 
-	if (about) {
-		gtk_window_present (GTK_WINDOW (about));
-		return;
-	}
+	priv = app->priv;
 	
-	about = gossip_about_new ();
-	g_object_add_weak_pointer (G_OBJECT (about), (gpointer) &about);
-	gtk_widget_show (about);
+	gossip_about_dialog_new (GTK_WINDOW (priv->window));
 }
 
 static void
@@ -985,6 +1005,7 @@ app_session_protocol_connected_cb (GossipSession  *session,
 	priv = app->priv;
 
 	app_connection_items_update ();
+	app_accounts_update_toolbar ();
 	app_presence_updated ();
 }
 
@@ -995,13 +1016,28 @@ app_session_protocol_disconnected_cb (GossipSession  *session,
 				      gpointer        user_data)
 {
 	GossipAppPriv *priv;
-	
+
 	priv = app->priv;
-	
+
  	app_connection_items_update ();
+	app_accounts_update_toolbar ();
 	app_presence_updated ();
 }
 
+static void
+app_session_protocol_error_cb (GossipSession  *session,
+			       GossipProtocol *protocol,
+			       GossipAccount  *account,
+			       GError         *error,
+			       gpointer        user_data)
+{
+	GossipAppPriv *priv;
+
+	priv = app->priv;
+
+	gtk_widget_show (priv->accounts_toolbar);
+}
+
 static gchar * 
 app_session_get_password_cb (GossipSession *session,
 			     GossipAccount *account,
@@ -1215,7 +1251,7 @@ app_tray_create (void)
 	priv = app->priv;
 
 	priv->tray_icon = egg_tray_icon_new (_("Gossip, Instant Messaging Client"));
-		
+
 	priv->tray_event_box = gtk_event_box_new ();
 	priv->tray_image = gtk_image_new ();
 	
@@ -1248,6 +1284,10 @@ app_tray_create (void)
 			  "destroy",
 			  G_CALLBACK (app_tray_destroy_cb),
 			  priv->tray_event_box);
+
+#ifdef HAVE_LIBNOTIFY
+	gossip_notify_set_attach_widget (GTK_WIDGET (priv->tray_icon));
+#endif
 }
 
 static void
@@ -1431,7 +1471,8 @@ app_idle_check_cb (GossipApp *app)
  *
  */
 void
-gossip_app_connect (gboolean startup)
+gossip_app_connect (GossipAccount *account,
+		    gboolean       startup)
 {
 	GossipAppPriv        *priv;
 	GossipAccountManager *manager;
@@ -1462,7 +1503,7 @@ gossip_app_connect (gboolean startup)
 		return;
 	}
 	
-	gossip_session_connect (priv->session, NULL, startup);
+	gossip_session_connect (priv->session, account, startup);
 }
 
 /* Test hack to add support for disconnecting all accounts and then connect 
@@ -1546,7 +1587,6 @@ app_disconnect (void)
 	GossipAppPriv *priv = app->priv;
 
 	gossip_session_disconnect (priv->session, NULL);
-	app_status_flash_stop ();
 }
 
 static void
@@ -1613,6 +1653,10 @@ app_connection_items_update (void)
 	for (l = priv->widgets_disconnected; l; l = l->next) {
 		gtk_widget_set_sensitive (l->data, (disconnected > 0));
 	}
+
+	if (connected == 0) {
+		app_status_flash_stop ();
+	}
 }
 
 gboolean
@@ -1740,7 +1784,7 @@ app_accounts_create (void)
 
 	g_list_free (accounts);
 
-	/* Show/hide toolbar */
+	/* Update separator position */
 	app_accounts_update_separator ();
 }
 
@@ -1757,7 +1801,7 @@ app_accounts_update_separator (void)
 
 	DEBUG_MSG (("AppAccounts: Updating toolbar separator position"));
 
-	/* remove current separator */
+	/* Remove current separator */
 	children = gtk_container_get_children (GTK_CONTAINER (priv->accounts_toolbar));
 	for (l = children; l; l = l->next) {
 		if (G_OBJECT_TYPE (l->data) == GOSSIP_TYPE_ACCOUNT_BUTTON) {
@@ -1769,23 +1813,23 @@ app_accounts_update_separator (void)
 	
 	g_list_free (children);
 
-	/* find position for new separator */
+	/* Find position for new separator */
 	children = gtk_container_get_children (GTK_CONTAINER (priv->accounts_toolbar));
 	for (l = children; l; l = l->next) {
 		GossipAccountButton *account_button;
 
 		account_button = l->data;
 
-		if (gossip_account_button_get_is_important (account_button)) {
+		if (gossip_account_button_is_important (account_button)) {
 			index++;
 		} else {
 			break;
 		}
 	}
 
-	/* don't put a separator in if it is at the start or end */
+	/* Don't put a separator in if it is at the start or end */
 	if (index > 0 && index < g_list_length (children)) {
-		/* create separator */
+		/* Create separator */
 		item = gtk_separator_tool_item_new ();
 		gtk_toolbar_insert (GTK_TOOLBAR (priv->accounts_toolbar), item, index);
 		gtk_widget_show (GTK_WIDGET (item));
@@ -1797,23 +1841,42 @@ app_accounts_update_separator (void)
 static void 
 app_accounts_update_toolbar (void)
 {
-	GossipAppPriv        *priv;
-	GossipAccountManager *manager;
-	gint                  count;
+	GossipAppPriv *priv;
+	GList         *children;
+	GList         *l;
+	gint           count;
+	gboolean       errors;              
 
 	priv = app->priv;
 
-	manager = gossip_session_get_account_manager (priv->session);
-	count = gossip_account_manager_get_count (manager);
+	/* Find position for new separator */
+	children = gtk_container_get_children (GTK_CONTAINER (priv->accounts_toolbar));
+	for (l = children, errors = FALSE; l && !errors; l = l->next) {
+		GossipAccountButton *account_button;
 
-	DEBUG_MSG (("AppAccounts: Updating toolbar"));
+		if (G_OBJECT_TYPE (l->data) != GOSSIP_TYPE_ACCOUNT_BUTTON) {
+			continue;
+		}
 
-	/* show accounts if we have more than one */
-	if (count < 2) {
+		account_button = l->data;
+
+		if (gossip_account_button_is_error_shown (account_button)) {
+			errors = TRUE;
+		}
+	}
+
+	count = g_list_length (children);
+
+	/* Show accounts if we have more than one */
+	if (count < 2 || errors) {
 		gtk_widget_hide (priv->accounts_toolbar);
 	} else {
 		gtk_widget_show (priv->accounts_toolbar);
 	}
+
+	DEBUG_MSG (("AppAccounts: There are %d accounts on the toolbar", count));
+
+	g_list_free (children);
 }
 
 static void
@@ -1857,24 +1920,33 @@ static void
 app_accounts_remove (GossipAccount *account)
 {
 	GossipAppPriv *priv;
+	GtkContainer  *container;
 	GList         *children;
 	GList         *l;
+	gboolean       remove_separator;
 
 	priv = app->priv;
 
 	DEBUG_MSG (("AppAccounts: Removing account with id:'%s'", 
 		   gossip_account_get_id (account)));
 
-	children = gtk_container_get_children (GTK_CONTAINER (priv->accounts_toolbar));
+	container = GTK_CONTAINER (priv->accounts_toolbar);
+	children = gtk_container_get_children (container);
+	remove_separator = g_list_length (children) == 3;
+
 	for (l = children; l; l = l->next) {
 		GossipAccount *this_account;
 
 		if (G_OBJECT_TYPE (l->data) != GOSSIP_TYPE_ACCOUNT_BUTTON) {
+			if (remove_separator) {
+				gtk_container_remove (container, l->data);
+			}
+
 			continue;
 		}
 		
 		this_account = gossip_account_button_get_account (l->data);
-		if (! gossip_account_equal (account, this_account)) {
+		if (!gossip_account_equal (account, this_account)) {
 			continue;
 		}
 
@@ -1882,8 +1954,7 @@ app_accounts_remove (GossipAccount *account)
 						      G_CALLBACK (app_accounts_account_notify_cb), 
 						      NULL);
 		
-		gtk_container_remove (GTK_CONTAINER (priv->accounts_toolbar), 
-				      l->data);
+		gtk_container_remove (container, l->data);
 	}
 	
 	g_list_free (children);
@@ -2210,9 +2281,11 @@ app_chatroom_auto_connect_update_cb (GossipChatroomManager    *manager,
 				     GossipChatroomJoinResult  result,
 				     gpointer                  user_data)
 {
-	const gchar *chatroom_name;
+	GossipChatroom *chatroom;
+	const gchar    *name;
 
-	chatroom_name = gossip_chatroom_provider_get_room_name (provider, id);
+	chatroom = gossip_chatroom_provider_find (provider, id);
+	name = gossip_chatroom_get_name (chatroom);
 
 	switch (result) {
 	case GOSSIP_CHATROOM_JOIN_NICK_IN_USE:
@@ -2229,7 +2302,8 @@ app_chatroom_auto_connect_update_cb (GossipChatroomManager    *manager,
 	case GOSSIP_CHATROOM_JOIN_ALREADY_OPEN:
 		DEBUG_MSG (("AppChatroom: Auto connect update: success for room:'%s'",
 			    chatroom_name));
-		gossip_group_chat_show (provider, id);
+
+		gossip_group_chat_new (provider, id);
 		break;
 
  	case GOSSIP_CHATROOM_JOIN_CANCELED:
